(declare
 ;; This wants to be an alternative to hash tables.  Thus our updates
 ;; must share the atomicity properties.
 (disable-interrupts)

 (no-argc-checks)
 (no-bound-checks)
 (no-procedure-checks)
 (safe-globals)
 (specialize)
 )

(module
 llrb-string-table
 (
  make-binding-set
  empty-binding-set
  binding-set-empty?
  binding-set-ref/default
  binding-set-ref
  binding-set-insert
  binding-set-update
  binding-set-cons
  binding-set-fold
  binding-set-union
  ;;
  make-string-table
  string-table?
  string-table-copy
  string-table-empty?
  string-table-delete!
  string-table-set!
  string-table-ref/default
  string-table-ref
  string-table-update!
  string-table-fold
  string-table-for-each
  string-table/min
  string-table/delete-min!
  ;;
  str2sym ;; caches string->symbol
  )
 (import scheme chicken)
 (import (only data-structures identity))
 (import llrb-syntax)
 ;;(include "llrbsyn.scm")

 (define-syntax typecheck
   (syntax-rules ()
     ((_ obj typetag loc) (##sys#check-structure obj typetag loc))))

 (define-syntax checkbinding-node
   (syntax-rules ()
     ((_ obj loc) (typecheck obj '<string-binding-node> loc))))
 
 (define (make-binding-node color left right key value)
   (##sys#make-structure '<string-binding-node> color left right key value))
 (define (binding-node? obj)
   (##sys#structure? obj '<string-binding-node>))
 (define-syntax binding-node-color
   (syntax-rules ()
     ((_ n) (##sys#slot n 1))))
 (define-syntax binding-node-left
   (syntax-rules ()
     ((_ n) (##sys#slot n 2))))
 (define-syntax binding-node-right
   (syntax-rules ()
     ((_ n) (##sys#slot n 3))))
 (define-syntax binding-node-key
   (syntax-rules ()
     ((_ n) (##sys#slot n 4))))
 (define-syntax binding-node-value
   (syntax-rules ()
     ((_ n) (##sys#slot n 5))))

 (define-syntax binding-set-update-node
   (syntax-rules (left: right: color:)
     ((_ 1 n l r c ())
      (make-binding-node c l r (binding-node-key n) (binding-node-value n)))
     ((_ 1 n l r c (left: v . more))
      (binding-set-update-node 1 n v r c more))
     ((_ 1 n l r c (right: v . more))
      (binding-set-update-node 1 n l v c more))
     ((_ 1 n l r c (color: v . more))
      (binding-set-update-node 1 n l r v more))
     ((_ n . more)
      (binding-set-update-node
       1 n (binding-node-left n) (binding-node-right n) (binding-node-color n) more))))
 
 (define-syntax %string<?
   (syntax-rules ()
     ((_ s1 s2)
      (let ((len1 (##core#inline "C_block_size" s1))
	    (len2 (##core#inline "C_block_size" s2)))
	(let ((cmp (##core#inline
		    "C_string_compare" s1 s2
		    (if (fx< len1 len2) len1 len2) )))
	  (or (fx< cmp 0)
	      (and (fx< len1 len2)
		   (eq? cmp 0) ) ))))))

 (define-syntax %string=?
   (syntax-rules ()
     ((_ s1 s2)
      (let ((len1 (##core#inline "C_block_size" s1))
	    (len2 (##core#inline "C_block_size" s2)))
	(and (eq? len1 len2)
	     (eq?
	      (##core#inline
	       "C_string_compare" s1 s2
	       (if (fx< len1 len2) len1 len2) )
	      0))))))

 (define-syntax binding-node-key-node-eq?
   (syntax-rules ()
     ((_ key node) (%string=? key (binding-node-key node)))))

 (define-syntax binding-node-key-node-lt
   (syntax-rules () ((_ key node) (%string<? key (binding-node-key node)))))

 (define-syntax binding-node-node-node-=?
   (syntax-rules ()
     ((_ node1 node2) (%string=? (binding-node-key node1) (binding-node-key node2)))))

 (define-syntax binding-node-node-node-lt
   (syntax-rules ()
     ((_ node1 node2) (%string<? (binding-node-key node1) (binding-node-key node2)))))

 (define-llrbtree/positional
   (ordered pure)
   binding-set-update-node
   binding-node-init!		 ;; defined
   binding-node-lookup		 ;; defined
   binding-node-min		 ;; defined
   binding-node-fold		 ;; defined
   binding-node-for-each	 ;; defined
   binding-node-insert		 ;; defined
   binding-node-delete		 ;; defined
   binding-node-delete-min	 ;; defined
   binding-node-empty?		 ;; defined
   binding-node-key-node-eq?
   binding-node-node-node-=?
   binding-node-key-node-lt
   binding-node-node-node-lt     ;; before? node order
   binding-node-left
   binding-node-right
   binding-node-color
   )

 ;; Constructors

 ;; 0X0

 (define (%make-new-binding-node k v)	; internal/unclean
   (make-binding-node #f #f #f k v))

 (define empty-binding-set
   (let ((empty-node (binding-node-init! (make-binding-node #f #f #f #f #f))))
     (lambda () empty-node)))		; export

 ;; 0Xpairs

 (: make-binding-set (&rest -> (struct <string-binding-node>)))
 (define (make-binding-set . lst)	; export
   (if (null? lst)
       (empty-binding-set)
       (do ((lst lst (cdr lst))
	    (set (empty-binding-set)
		 (let* ((x (car lst)) (k (car x)))
		   (binding-node-insert set k #f (%make-new-binding-node k (cdr x)) #f))))
	   ((null? lst) set))))

 (define (%binding-set-ref/thunk envt k thunk) ; internal
   (let ((entry (binding-node-lookup envt k)))
     (if entry (binding-node-value entry) (thunk))))

 (define (%binding-set-ref/default envt k default) ; internal
   (let ((entry (binding-node-lookup envt k)))
     (if entry (binding-node-value entry) default)))

 (: binding-set-ref/default ((struct <string-binding-node>) string * --> *))
 (define (binding-set-ref/default envt k default) ; export
   (checkbinding-node envt 'binding-set-ref/default)
   (%binding-set-ref/default envt k default))

 (: binding-set-ref ((struct <string-binding-node>) string &rest (procedure () . *) --> *))
 (define (binding-set-ref envt k . thunk) ; export
   (checkbinding-node envt 'binding-set-ref)
   (%binding-set-ref/thunk
    envt k
    (if (pair? thunk) (car thunk)
	(lambda ()
	  (error "binding-set-ref unbound key" k)))))

 ;; setXkeyXvalue

 (: binding-set-insert ((struct <string-binding-node>) * * --> (struct <string-binding-node>)))
 (define (binding-set-insert nodeset k v) ; export
   (checkbinding-node nodeset 'binding-set-insert)
   (binding-node-insert nodeset k #f (%make-new-binding-node k v) #f))

 (: binding-set-update ((struct <string-binding-node>)
			string (procedure (*) *) (procedure () *)
			--> (struct <string-binding-node>)))
 (define (binding-set-update nodeset k update dflt) ; export
   (checkbinding-node nodeset 'binding-set-update)
   (ensure procedure? update)
   (ensure procedure? dflt)
   (binding-node-insert
    nodeset k #f
    (let ((update (or update identity)))
      (lambda (n)
	(let ((v (update (binding-node-value n))))
	  (make-binding-node #f #f #f (binding-node-key n) v))))
    (lambda () (%make-new-binding-node k (dflt)))))

 ;; srfi-1::alist-cons compatible
 (: binding-set-cons (string * (struct <string-binding-node>) --> (struct <string-binding-node>)))
 (define (binding-set-cons k v nodeset) ; export
   (binding-set-insert nodeset k v))

 (: binding-set-fold ((struct <string-binding-node>) (procedure (string * *) . *) * -> *))
 (define (binding-set-fold nodeset kvcons nil)
   (checkbinding-node nodeset 'binding-set-fold)
   (ensure procedure? kvcons)
   (binding-node-fold (lambda (e i) (kvcons (binding-node-key e) (binding-node-value e) i)) nil nodeset))

 ;; setXset

 (: binding-set-union ((struct <string-binding-node>) (struct <string-binding-node>) --> (struct <string-binding-node>)))
 (define (binding-set-union inner outer) ; export
   (checkbinding-node inner 'binding-set-union)
   (checkbinding-node outer 'binding-set-union)
   (binding-node-fold
    (lambda (node init) (binding-node-insert init (binding-node-key node) #f node #f))
    outer inner))

 (define-type :table: (struct <llrb-string-table>))
 (define-record-type <llrb-string-table>
   (%make-string-table root)
   string-table?
   (root root root-set!))

 (define-syntax check-table
   (syntax-rules ()
     ((_ obj loc) (typecheck obj '<llrb-string-table> loc))))

 (define (make-string-table)
   (%make-string-table (empty-binding-set)))

 (: string-table-copy (:table: --> :table:))
 (define (string-table-copy table)
   (%make-string-table (root table)))

 (: string-table-empty? (:table: --> boolean))
 (define (string-table-empty? table)
   (check-table table 'string-table-empty?)
   (binding-node-empty? (root table) (root table)))

 (: string-table-delete! (:table: string -> *))
 (define (string-table-delete! table key)
   (check-table table 'string-table-delete!)
   (root-set! table (binding-node-delete (root table) key)))

 (: string-table-set! (:table: string * -> *))
 (define (string-table-set! table key value)
   (check-table table 'string-table-set!)
   (root-set! table (binding-node-insert (root table) key #f (%make-new-binding-node key value) #f)))
 
 (: string-table-ref/default (:table: string * --> *))
 (define (string-table-ref/default table key default)
   (check-table table 'string-table-ref/default)
   (%binding-set-ref/default (root table) key default))

 (: string-table-ref (:table: string &rest (procedure () *) -> *))
 (define (string-table-ref table key . thunk)
   (check-table table 'string-table-ref)
   (%binding-set-ref/thunk
    (root table) key
    (if (pair? thunk) (car thunk)
	(lambda ()
	  (error "string-table-ref unbound key" key)))))

 (: string-table-update! (:table: string (or false procedure) &rest procedure -> *))
 (define (string-table-update! table key update . default)
   (check-table table 'string-table-update!)
   (or (eq? update #f) (ensure procedure? update))
   (let ((result #f))
     (root-set!
      table
      (binding-node-insert
       (root table)
       key
       #f
       (if update
	   (lambda (n)
	     (let ((v (update (binding-node-value n))))
	       (set! result v)
	       (make-binding-node #f #f #f (binding-node-key n) v)))
	   (lambda (n) (set! result (binding-node-value n)) n))
       (lambda ()
	 (if (and (pair? default) (procedure? (car default)))
	     (let ((thunk (car default))) (%make-new-binding-node key (thunk)))
	     (error "string-table-update! no default" default)))))
     result))

 (: string-table-fold ((procedure (string * :table:) *) * :table: -> *))
 (define (string-table-fold proc init table)
   (check-table table 'string-table-fold)
   (ensure procedure? proc)
   (binding-node-fold
    (lambda (node init) (proc (binding-node-key node) (binding-node-value node) init))
    init (root table)))

 (: string-table-for-each ((procedure (string *) *) :table: -> *))
 (define (string-table-for-each proc table)
   (check-table table 'string-table-for-each)
   (ensure procedure? proc)
   (binding-node-for-each
    (lambda (node) (proc (binding-node-key node) (binding-node-value node)))
    (root table))
   #f)

 (: string-table/min (:table: (procedure () * *) --> * *))
 (define (string-table/min table default)
   (check-table table 'string-table/min)
   (let ((node (binding-node-min (root table))))
     (if node (values (binding-node-key node) (binding-node-value node))
	 (begin
	   (ensure procedure? default)
	   (default)))))

 (: string-table/delete-min! (:table: (procedure () * *) -> * *))
 (define (string-table/delete-min! table default)
   (if (string-table-empty? table) ;; checks the table type.
       (begin
	 (ensure procedure? default)
	 (default))
       (binding-node-delete-min
	(root table)
	(lambda (root node)
	  (root-set! table root)
	  (values (binding-node-key node) (binding-node-value node))))))

 (: str2sym (string -> symbol))
 (define str2sym
   (let ((set (empty-binding-set))
	 (string->symbol string->symbol))
     (lambda (s)
       (ensure string? s)
       (or (%binding-set-ref/default set s #f)
	   (let ((value (string->symbol s)))
	     (set! set (binding-node-insert set s #f (%make-new-binding-node s value) #f))
	     value)))))

 )

(module
 llrb-fixnum-table
 (
  make-fixnum-table
  fixnum-table-copy
  fixnum-table-ref
  fixnum-table-ref/default
  fixnum-table-set!
  fixnum-table-delete!
  fixnum-table-for-each
  fixnum-table-fold
  fixnum-table-update!
  fixnum-table/min fixnum-table/delete-min!
  )
 (import scheme chicken)
 (import (only data-structures identity))
 (import llrb-syntax)
 ;;(include "llrbsyn.scm")

 (define-syntax typecheck
   (syntax-rules ()
     ((_ obj typetag loc) (##sys#check-structure obj typetag loc))))

 (define-syntax checkbinding-node
   (syntax-rules ()
     ((_ obj loc) (typecheck obj '<fixnum-binding-node> loc))))

 (define (make-binding-node color left right key value)
   (##sys#make-structure '<fixnum-binding-node> color left right key value))
 (define (binding-node? obj)
   (##sys#structure? obj '<fixnum-binding-node>))
 (define-syntax binding-node-color
   (syntax-rules ()
     ((_ n) (##sys#slot n 1))))
 (define-syntax binding-node-left
   (syntax-rules ()
     ((_ n) (##sys#slot n 2))))
 (define-syntax binding-node-right
   (syntax-rules ()
     ((_ n) (##sys#slot n 3))))
 (define-syntax binding-node-key
   (syntax-rules ()
     ((_ n) (##sys#slot n 4))))
 (define-syntax binding-node-value
   (syntax-rules ()
     ((_ n) (##sys#slot n 5))))

 (define-type :table: (struct <llrb-fixnum-table>))
 (define-record-type <llrb-fixnum-table>
   (%make-fixnum-table root)
   fixnum-table?
   (root root root-set!))

 (define-syntax check-table
   (syntax-rules ()
     ((_ obj loc) (typecheck obj '<llrb-fixnum-table> loc))))

 (define-syntax binding-set-update
   (syntax-rules (left: right: color:)
     ((_ 1 n l r c ())
      (make-binding-node c l r (binding-node-key n) (binding-node-value n)))
     ((_ 1 n l r c (left: v . more))
      (binding-set-update 1 n v r c more))
     ((_ 1 n l r c (right: v . more))
      (binding-set-update 1 n l v c more))
     ((_ 1 n l r c (color: v . more))
      (binding-set-update 1 n l r v more))
     ((_ n . more)
      (binding-set-update
       1 n (binding-node-left n) (binding-node-right n) (binding-node-color n) more))))

 (define-syntax binding-node-key-node-eq?
   (syntax-rules ()
     ((_ key node) (eq? key (binding-node-key node)))))

 (define-syntax binding-node-key-node-lt
   (syntax-rules () ((_ key node) (fx< key (binding-node-key node)))))

 (define-syntax binding-node-node-node-=?
   (syntax-rules ()
     ((_ node1 node2) (eq? (binding-node-key node1) (binding-node-key node2)))))

 (define-syntax binding-node-node-node-lt
   (syntax-rules ()
     ((_ node1 node2) (fx< (binding-node-key node1) (binding-node-key node2)))))

 (define-llrbtree/positional
   (ordered pure)
   binding-set-update
   binding-node-init!		 ;; defined
   binding-node-lookup		 ;; defined
   binding-node-min		 ;; defined
   binding-node-fold		 ;; defined
   binding-node-for-each	 ;; defined
   binding-node-insert		 ;; defined
   binding-node-delete		 ;; defined
   binding-node-delete-min	 ;; defined
   binding-node-empty?		 ;; defined
   binding-node-key-node-eq?
   binding-node-node-node-=?
   binding-node-key-node-lt
   binding-node-node-node-lt     ;; before? node order
   binding-node-left
   binding-node-right
   binding-node-color
   )

 (: make-fixnum-table (--> :table:))

 (define make-fixnum-table
   (let ((n0 (binding-node-init! (make-binding-node #f #f #f  #f #f))))
     (lambda () (%make-fixnum-table n0))))

 (: fixnum-table-copy (:table: --> :table:))
 (define (fixnum-table-copy table)
   (check-table table 'fixnum-table-copy)
   (%make-fixnum-table (root table)))

 (: fixnum-table-empty? (:table: --> boolean))
 (define (fixnum-table-empty? table)
   (ensure fixnum-table? table)
   (binding-node-empty? (root table) (root table)))

 (: fixnum-table-delete! (:table: fixnum -> *))
 (define (fixnum-table-delete! table key)
   (check-table table 'fixnum-table-delete!)
   (ensure fixnum? key)
   (root-set! table (binding-node-delete (root table) key)))

 (: fixnum-table-set! (:table: fixnum * -> boolean))
 (define (fixnum-table-set! table key value)
   (check-table table 'fixnum-table-set!)
   (ensure fixnum? key)
   (root-set! table (binding-node-insert (root table) key #f (make-binding-node #f #f #f key value) #f)))

 (: fixnum-table-ref/default (:table: fixnum * --> *))
 (define (fixnum-table-ref/default table key default)
   (check-table table 'fixnum-table-ref/default)
   (ensure fixnum? key)
   (let ((node (binding-node-lookup (root table) key)))
     (if node (binding-node-value node) default)))

 (: fixnum-table-ref (:table: fixnum &rest (procedure () *) -> *))
 (define (fixnum-table-ref table key . default)
   (check-table table 'fixnum-table-ref)
   (ensure fixnum? key)
   (let ((node (binding-node-lookup (root table) key)))
     (if node (binding-node-value node)
	 (if (pair? default) ((car default))
	     (error "fixnum-table-ref no key" key)))))

 (: fixnum-table-update! (:table: fixnum procedure &rest procedure -> *))
 (define (fixnum-table-update! table key update . default)
   (check-table table 'fixnum-table-update!)
   (ensure fixnum? key)
   (ensure procedure? update)
   (let ((result #f))
     (root-set!
      table
      (binding-node-insert
       (root table) key #f
       (let ((update (or update identity)))
	 (lambda (n)
	   (let ((v (update (binding-node-value n))))
	     (set! result v)
	     (make-binding-node #f #f #f (binding-node-key n) v))))
       (if (and (pair? default) (procedure? (car default)))
	   (let ((thunk (car default)))
	     (lambda () (make-binding-node #f #f #f key (thunk))))
	   (lambda () (error "fixnum-table-update! no default" default)))))
     result))

 (: fixnum-table-fold ((procedure (fixnum * *) *) * :table: -> *))
 (define (fixnum-table-fold proc init table)
   (check-table table 'fixnum-table-fold)
   (ensure procedure? proc)
   (binding-node-fold
    (lambda (node init) (proc (binding-node-key node) (binding-node-value node) init))
    init (root table)))

 (: fixnum-table-for-each ((procedure (fixnum *) *) :table: -> *))
 (define (fixnum-table-for-each proc table)
   (check-table table 'fixnum-table-for-each)
   (ensure procedure? proc)
   (binding-node-for-each
    (lambda (node) (proc (binding-node-key node) (binding-node-value node)))
    (root table))
   #f)

 (: fixnum-table/min (:table: (procedure () * *) --> * *))
 (define (fixnum-table/min table default)
   (check-table table 'fixnum-table/min)
   (let ((node (binding-node-min (root table))))
     (if node (values (binding-node-key node) (binding-node-value node))
	 (begin
	   (ensure procedure? default)
	   (default)))))

 (: fixnum-table/delete-min! (:table: (procedure () fixnum *) -> * *))
 (define (fixnum-table/delete-min! table default)
   (if (fixnum-table-empty? table) ;; dos the type check
       (begin
	 (ensure procedure? default)
	 (default))
       (binding-node-delete-min
	(root table)
	(lambda (root node)
	  (root-set! table root)
	  (values (binding-node-key node) (binding-node-value node))))))

 
 )

(module
 llrb-symbol-tree
 (
  make-binding-set
  empty-binding-set
  binding-set-empty?
  binding-set-ref/default
  binding-set-ref
  binding-set-insert
  binding-set-update
  binding-set-cons
  binding-set-fold
  binding-set-union
  ;;
  make-symbol-table
  symbol-table-copy
  symbol-table?
  symbol-table-set!
  symbol-table-ref/default
  symbol-table-ref
  symbol-table-update!
  )
 (import scheme chicken)
 (import (only data-structures identity))
 (import llrb-syntax)
 ;;(include "llrbsyn.scm")

 (define-syntax typecheck
   (syntax-rules ()
     ((_ obj typetag loc) (##sys#check-structure obj typetag loc))))

 (define-syntax checkbinding-node
   (syntax-rules ()
     ((_ obj loc) (typecheck obj '<symbol-binding-node> loc))))

 (define (make-binding-node color left right key value)
   (##sys#make-structure '<symbol-binding-node> color left right key value))
 (define (binding-node? obj)
   (##sys#structure? obj '<symbol-binding-node>))
 (define-syntax binding-node-color
   (syntax-rules ()
     ((_ n) (##sys#slot n 1))))
 (define-syntax binding-node-left
   (syntax-rules ()
     ((_ n) (##sys#slot n 2))))
 (define-syntax binding-node-right
   (syntax-rules ()
     ((_ n) (##sys#slot n 3))))
 (define-syntax %binding-node-name
   (syntax-rules ()
     ((_ n) (##sys#slot n 4))))
 (define (binding-node-name n)
   (checkbinding-node n 'binding-node-name)
   (%binding-node-name n))
 (define-syntax %binding-node-value
   (syntax-rules ()
     ((_ n) (##sys#slot n 5))))
 (define (binding-node-value n)
   (checkbinding-node n 'binding-node-value)
   (%binding-node-value n))

 (define-syntax %symbol->string
   (syntax-rules ()
     ((_ s) (##sys#symbol->string s))))

 (define-syntax %string<?
   (syntax-rules ()
     ((_ s1 s2)
      (let ((len1 (##core#inline "C_block_size" s1))
	    (len2 (##core#inline "C_block_size" s2)))
	(let ((cmp (##core#inline
		    "C_string_compare" s1 s2
		    (if (fx< len1 len2) len1 len2) )))
	  (or (fx< cmp 0)
	      (and (fx< len1 len2)
		   (eq? cmp 0) ) ))))))

 (define-syntax binding-set-update-node
   (syntax-rules (left: right: color:)
     ((_ 1 n l r c ())
      (make-binding-node c l r (%binding-node-name n) (%binding-node-value n)))
     ((_ 1 n l r c (left: v . more))
      (binding-set-update-node 1 n v r c more))
     ((_ 1 n l r c (right: v . more))
      (binding-set-update-node 1 n l v c more))
     ((_ 1 n l r c (color: v . more))
      (binding-set-update-node 1 n l r v more))
     ((_ n . more)
      (binding-set-update-node
       1 n (binding-node-left n) (binding-node-right n) (binding-node-color n) more))))

 (define-syntax binding-k-n-eq?
   (syntax-rules () ((_ k n) (eq? k (%binding-node-name n)))))

 (define-syntax binding-n-n-eq?
   (syntax-rules () ((_ n1 n2) (eq? (%binding-node-name n1) (%binding-node-name n2)))))

 (define-syntax binding-k-n-lt
   (syntax-rules () ((_ k n) (%string<? k (%binding-node-name n)))))

 (define-syntax binding-n-n-lt
   (syntax-rules () ((_ node1 node2) (%string<? (%binding-node-name node1)
						(%binding-node-name node2)))))

 (define-llrbtree/positional
   (ordered pure)
   binding-set-update-node
   binding-set-init!    ;; defined
   binding-set-lookup   ;; defined
   #f			   ;; no min defined
   %binding-set-fold	   ;; defined
   #f			   ;; no for-each defined
   %binding-set-insert   ;; defined
   binding-set-delete   ;; defined
   #f			   ;; no delete-min defined
   %binding-set-empty?   ;; defined
   binding-k-n-eq?
   binding-n-n-eq?
   binding-k-n-lt
   binding-n-n-lt
   binding-node-left
   binding-node-right
   binding-node-color
   )

 (define (binding-set-empty? x)
   (ensure binding-node? x)
   (%binding-set-empty? x x))

 ;; Constructors

 ;; 0X0

 (define (%make-new-binding-node k v)	; internal/unclean
   (make-binding-node #f #f #f k v))

 (define %empty-binding-set		; internal
   (binding-set-init! (make-binding-node #f #f #f #f #f)))

 (define (empty-binding-set) %empty-binding-set)	; export

 ;; 0Xpairs

 (: make-binding-set (&rest -> (struct <symbol-binding-node>)))
 (define (make-binding-set . lst)	; export
   (if (null? lst)
       (empty-binding-set)
       (do ((lst lst (cdr lst))
	    (set %empty-binding-set
		 (let* ((x (car lst)) (k (%symbol->string (car x))))
		   (%binding-set-insert
		    set k #f (%make-new-binding-node k (cdr x)) #f))))
	   ((null? lst) set))))

 (define (%binding-set-ref/thunk envt k thunk) ; export?
   (let ((entry (binding-set-lookup envt k)))
       (if entry (%binding-node-value entry) (thunk))))

 (define (%binding-set-ref/default envt k default) ; internal
   (let ((entry (binding-set-lookup envt k)))
     (if entry (%binding-node-value entry) default)))

 (: binding-set-ref/default ((struct <symbol-binding-node>) symbol * --> *))
 (define (binding-set-ref/default envt k default) ; export
   (checkbinding-node envt 'binding-set-ref/default)
   (ensure symbol? k)
   (%binding-set-ref/default envt (%symbol->string k) default))

 (: binding-set-ref ((struct <symbol-binding-node>) symbol &rest (procedure () . *) --> *))
 (define (binding-set-ref envt k . thunk) ; export
   (checkbinding-node envt 'binding-set-ref)
   (ensure symbol? k)
   (%binding-set-ref/thunk
    envt (%symbol->string k)
    (if (pair? thunk) (car thunk)
	(lambda ()
	  (error "binding-set-ref unbound key" k)))))

 ;; setXkeyXvalue

 (: binding-set-insert ((struct <symbol-binding-node>) symbol * --> (struct <symbol-binding-node>)))
 (define (binding-set-insert nodeset k v) ; export
   (checkbinding-node nodeset 'binding-set-insert)
   (ensure binding-node? nodeset)
   (ensure symbol? k)
   (let ((k (%symbol->string k)))
     (%binding-set-insert nodeset k #f (%make-new-binding-node k v) #f)))

 (: binding-set-update ((struct <symbol-binding-node>) symbol (procedure (*) *) (procedure () *) --> (struct <symbol-binding-node>)))
 (define (binding-set-update nodeset k update dflt) ; export
   (checkbinding-node nodeset 'binding-set-update)
   (ensure symbol? k)
   (ensure procedure? update)
   (ensure procedure? dflt)
   (let ((k (%symbol->string k)))
     (%binding-set-insert
      nodeset k #f
      (lambda (n)
	(let ((v (update (binding-node-value n))))
	  (make-binding-node #f #f #f (%binding-node-name n) v)))
      (lambda () (%make-new-binding-node k (dflt))))))

 ;; srfi-1::alist-cons compatible
 (: binding-set-cons (symbol * (struct <symbol-binding-node>) --> (struct <symbol-binding-node>)))
 (define (binding-set-cons k v nodeset) ; export
   (binding-set-insert nodeset k v))

 (: binding-set-fold ((struct <symbol-binding-node>) (procedure (* *) . *) * -> *))
 (define (binding-set-fold nodeset kvcons nil)
   (checkbinding-node nodeset 'binding-set-fold)
   (ensure procedure? kvcons)
   (%binding-set-fold (lambda (e i) (kvcons (string->symbol (%binding-node-name e)) (%binding-node-value e) i)) nil nodeset))

 ;; setXset

 (: binding-set-union
    ((struct <symbol-binding-node>) (struct <symbol-binding-node>) --> (struct <symbol-binding-node>)))
 (define (binding-set-union inner outer) ; export
   (checkbinding-node inner 'binding-union)
   (checkbinding-node outer 'binding-union)
   (%binding-set-fold (lambda (node init) (%binding-set-insert init (%binding-node-name node) #f node #f)) outer inner))

 (define-type :table: (struct <llrb-symbol-table>))
 (define-record-type <llrb-symbol-table>
   (%make-symbol-table root)
   symbol-table?
   (root root root-set!))

 (define-syntax check-table
   (syntax-rules ()
     ((_ obj loc) (typecheck obj '<llrb-symbol-table> loc))))

 (define (make-symbol-table)
   (%make-symbol-table (empty-binding-set)))

 (: symbol-table-copy (:table: --> :table:))
 (define (symbol-table-copy table)
   (check-table table 'symbol-table-copy)
   (%make-symbol-table (root table)))

 (define (symbol-table-delete! table key)
   (check-table table 'symbol-table-delete!)
   (ensure symbol? key)
   (root-set! table (binding-set-delete (root table) (%symbol->string key))))

 (define (symbol-table-set! table key value)
   (check-table table 'symbol-table-set!)
   (ensure symbol? key)
   (let ((key (%symbol->string key)))
     (root-set! table (%binding-set-insert (root table) key #f (%make-new-binding-node key value) #f))))

 
 (: symbol-table-ref/default (:table: symbol * --> *))
 (define (symbol-table-ref/default table key default)
   (check-table table 'symbol-table-ref/default)
   (ensure symbol? key)
   (%binding-set-ref/default (root table) (%symbol->string key) default))

 (: symbol-table-ref (:table: symbol &rest (procedure () *) -> *))
 (define (symbol-table-ref table key . thunk)
   (check-table table 'symbol-table-ref)
   (ensure symbol? key)
   (%binding-set-ref/thunk
    (root table) (%symbol->string key)
    (if (pair? thunk) (car thunk)
	(lambda ()
	  (error "symbol-table-ref unbound key" key)))))

 (: symbol-table-update! (:table: symbol procedure &rest procedure -> *))
 (define (symbol-table-update! table key update . default)
   (check-table table 'symbol-table-update!)
   (ensure symbol? key)
   (ensure procedure? update)
   (let ((key (%symbol->string key))
	 (result #f))
     (root-set!
      table
      (%binding-set-insert
       (root table)
       key
       #f
       (let ((update (or update identity)))
	 (lambda (n)
	   (let ((v (update (binding-node-value n))))
	     (set! result v)
	     (make-binding-node #f #f #f (%binding-node-name n) v))))
       (if (and (pair? default) (procedure? (car default)))
	   (let ((thunk (car default))) (lambda () (%make-new-binding-node key (thunk))))
	   (lambda () (error "symbol-table-update! no default" default)))))
     result))
 )

(module
 llrb-generic-tree
 ;; FIXME: This should NOT be part of the same compilation unit as the
 ;; other (tuned) modules.  Here we REALLY should check procedure,
 ;; arguments etc.
 (
  make-llrb-treetype
  llrb-tree-type?
  make-binding-set
  empty-binding-set
  binding-set-empty?
  binding-set-ref/default
  binding-set-ref
  binding-set-insert
  binding-set-update
  binding-set-cons
  binding-set-fold
  binding-set-union
  ;;
  make-generic-table
  generic-table?
  generic-table-copy
  generic-table-empty?
  generic-table-delete!
  generic-table-set!
  generic-table-ref/default
  generic-table-ref
  generic-table-update!
  generic-table-fold
  generic-table-for-each
  generic-table/min
  generic-table/delete-min!
  )
 (import scheme chicken)
 (import (only data-structures identity))
 (import llrb-syntax)
 ;;(include "llrbsyn.scm")

 (define-syntax typecheck
   (syntax-rules ()
     ((_ obj typetag loc) (##sys#check-structure obj typetag loc))))

 (define-syntax checkbinding-node
   (syntax-rules ()
     ((_ obj loc) (typecheck obj '<binding-node> loc))))

 (define (make-binding-node color left right key value)
   (##sys#make-structure '<binding-node> color left right key value))
 (define (binding-node? obj)
   (##sys#structure? obj '<binding-node>))
 (define-syntax binding-node-color
   (syntax-rules ()
     ((_ n) (##sys#slot n 1))))
 (define-syntax binding-node-left
   (syntax-rules ()
     ((_ n) (##sys#slot n 2))))
 (define-syntax binding-node-right
   (syntax-rules ()
     ((_ n) (##sys#slot n 3))))
 (define-syntax binding-node-key
   (syntax-rules ()
     ((_ n) (##sys#slot n 4))))
 (define-syntax binding-node-value
   (syntax-rules ()
     ((_ n) (##sys#slot n 5))))

 (define-syntax binding-set-update-node
   (syntax-rules (left: right: color:)
     ((_ 1 n l r c ())
      (make-binding-node c l r (binding-node-key n) (binding-node-value n)))
     ((_ 1 n l r c (left: v . more))
      (binding-set-update-node 1 n v r c more))
     ((_ 1 n l r c (right: v . more))
      (binding-set-update-node 1 n l v c more))
     ((_ 1 n l r c (color: v . more))
      (binding-set-update-node 1 n l r v more))
     ((_ n . more)
      (binding-set-update-node
       1 n (binding-node-left n) (binding-node-right n) (binding-node-color n) more))))

 (define-type :table-type: (struct llrb-tree-type))
 (define-record llrb-tree-type
   lookup
   min
   fold
   for-each
   insert
   delete
   delete-min
   )

 (define binding-node-init! #f)
 (define binding-node-empty? #f)

 (define (make-llrb-treetype equal less)

   (define-syntax generic-k-n-eq?
     (syntax-rules () ((_ k n) (equal k (binding-node-key n)))))

   (define-syntax generic-n-n-eq?
     (syntax-rules () ((_ n1 n2) (equal (binding-node-key n1) (binding-node-key n2)))))

   (define-syntax generic-k-n-lt
     (syntax-rules () ((_ k n) (less k (binding-node-key n)))))

   (define-syntax generic-n-n-lt
     (syntax-rules () ((_ node1 node2) (less (binding-node-key node1)
					     (binding-node-key node2)))))

   (define-llrbtree/positional
     (ordered pure)
     binding-set-update-node
     %binding-node-init!     ;; defined
     binding-node-lookup     ;; defined
     binding-node-min	     ;; min defined
     binding-node-fold	     ;; defined
     binding-node-for-each   ;; defined
     binding-node-insert     ;; defined
     binding-node-delete     ;; defined
     binding-node-delete-min ;; defined
     %binding-node-empty?    ;; defined
     generic-k-n-eq?
     generic-n-n-eq?
     generic-k-n-lt
     generic-n-n-lt
     binding-node-left
     binding-node-right
     binding-node-color
     )

   (if (not binding-node-init!)
       (begin
	 (set! binding-node-init! %binding-node-init!)
	 (set! binding-node-empty? (lambda (x) (checkbinding-node x 'binding-node-empty?) (%binding-node-empty? x x)))))

   (make-llrb-tree-type
    binding-node-lookup
    binding-node-min
    binding-node-fold
    binding-node-for-each
    binding-node-insert
    binding-node-delete
    binding-node-delete-min
    ))

 ;; Constructors

 ;; 0X0

 (define (%make-new-binding-node k v)	; internal/unclean
   (make-binding-node #f #f #f k v))

 (define (empty-binding-set)		; export
   (binding-node-init! (make-binding-node #f #f #f #f #f)))

 ;; 0Xpairs

 (: make-binding-set (:table-type: &rest -> (struct <binding-node>)))
 (define (make-binding-set type . lst)	; export
   (if (null? lst)
       (empty-binding-set)
       (let ((insert (llrb-tree-type-insert type)))
	 (do ((lst lst (cdr lst))
	    (set (empty-binding-set)
		 (let* ((x (car lst)) (k (car x)))
		   (insert set k #f (%make-new-binding-node k (cdr x)) #f))))
	   ((null? lst) set)))))

 (define (%binding-set-ref/thunk type envt k thunk) ; internal
   (let ((entry ((llrb-tree-type-lookup type) envt k)))
       (if entry (binding-node-value entry) (thunk))))

 (define (%binding-set-ref/default type envt k default) ; internal
   (let ((entry ((llrb-tree-type-lookup type) envt k)))
     (if entry (binding-node-value entry) default)))

 (: binding-set-ref/default (:table-type: (struct <binding-node>) * * --> *))
 (define (binding-set-ref/default type envt k default) ; export
   (ensure llrb-tree-type? type)
   (checkbinding-node envt 'binding-set-ref/default)
   (%binding-set-ref/default type envt k default))

 (: binding-set-ref (:table-type: (struct <binding-node>) * &rest (procedure () . *) --> *))
 (define (binding-set-ref type envt k . thunk) ; export
   (ensure llrb-tree-type? type)
   (checkbinding-node envt 'binding-set-ref)
   (%binding-set-ref/thunk
    type envt k
    (if (pair? thunk) (car thunk)
	(lambda ()
	  (error "binding-set-ref unbound key" k)))))

 ;; setXkeyXvalue

 (: binding-set-insert (:table-type: (struct <binding-node>) * * --> (struct <binding-node>)))
 (define (binding-set-insert type nodeset k v) ; export
   (ensure llrb-tree-type? type)
   (checkbinding-node nodeset 'binding-set-insert)
   ((llrb-tree-type-insert type) nodeset k #f (%make-new-binding-node k v) #f))

 (: binding-set-update (:table-type: (struct <binding-node>)
			* (procedure (*) *) (procedure () *)
			--> (struct <binding-node>)))
 (define (binding-set-update type nodeset k update dflt) ; export
   (ensure llrb-tree-type? type)
   (checkbinding-node nodeset 'binding-set-update)
   (ensure procedure? update)
   (ensure procedure? dflt)
   ((llrb-tree-type-insert type)
    nodeset k #f
    (let ((update (or update identity)))
      (lambda (n)
	(let ((v (update (binding-node-value n))))
	  (make-binding-node #f #f #f (binding-node-key n) v))))
    (lambda () (%make-new-binding-node k (dflt)))))

 ;; srfi-1::alist-cons compatible -- bahaha!  No longer, passing the type that way is bad.
 (: binding-set-cons (:table-type: * * (struct <binding-node>) --> (struct <binding-node>)))
 (define (binding-set-cons type k v nodeset) ; export
   (binding-set-insert type nodeset k v))

 (: binding-set-fold (:table-type: (struct <binding-node>) (procedure (* *) . *) * -> *))
 (define (binding-set-fold type nodeset kvcons nil)
   (ensure llrb-tree-type? type)
   (checkbinding-node nodeset 'binding-set-fold)
   (ensure procedure? kvcons)
   ((llrb-tree-type-fold type) (lambda (e i) (kvcons (binding-node-key e) (binding-node-value e) i)) nil nodeset))

 ;; setXset

 (: binding-set-union
    (:table-type: (struct <binding-node>) (struct <binding-node>) --> (struct <binding-node>)))
 (define (binding-set-union type inner outer) ; export
   (ensure llrb-tree-type? type)
   (checkbinding-node inner 'binding-set-union)
   (checkbinding-node outer 'binding-set-union)
   ((llrb-tree-type-fold type)
    (let ((insert (llrb-tree-type-insert type)))
      (lambda (node init) (insert init (binding-node-key node) #f node #f)))
    outer inner))

 (define-type :table: (struct <llrb-generic-table>))
 (define-record-type <llrb-generic-table>
   (%make-generic-table type root)
   generic-table?
   (type llrb-type)
   (root root root-set!))

 (define-syntax check-table
   (syntax-rules ()
     ((_ obj loc) (typecheck obj '<llrb-generic-table> loc))))

 (define (make-generic-table type)
   (ensure llrb-tree-type? type)
   (%make-generic-table type (empty-binding-set)))

 (: generic-table-copy (:table: --> :table:))
 (define (generic-table-copy table)
   (check-table table 'generic-table-copy)
   (%make-generic-table (llrb-type table) (root table)))

 (: generic-table-empty? (:table: --> boolean))
 (define (generic-table-empty? table)
   (check-table table 'generic-table-empty?)
   (binding-node-empty? (root table) (root table)))

 (define (generic-table-delete! table key)
   (check-table table 'generic-table-delete!)
   (root-set! table ((llrb-tree-type-delete (llrb-type table)) (root table) key)))

 (define (generic-table-set! table key value)
   (check-table table 'generic-table-set!)
   (root-set! table ((llrb-tree-type-insert (llrb-type table)) (root table) key #f (%make-new-binding-node key value) #f)))
 
 (: generic-table-ref/default (:table: * * --> *))
 (define (generic-table-ref/default table key default)
   (check-table table 'generic-table-ref/default)
   (%binding-set-ref/default (llrb-type table) (root table) key default))

 (: generic-table-ref (:table: * &rest (procedure () *) -> *))
 (define (generic-table-ref table key . thunk)
   (check-table table 'generic-table-ref)
   (%binding-set-ref/thunk
    (llrb-type table) (root table) key
    (if (pair? thunk) (car thunk)
	(lambda ()
	  (error "generic-table-ref unbound key" key)))))

 (: generic-table-update! (:table: * (or false procedure) &rest procedure -> *))
 (define (generic-table-update! table key update . default)
   (check-table table 'generic-table-update!)
   (or (eq? update #f) (ensure procedure? update))
   (let ((result #f))
     (root-set!
      table
      ((llrb-tree-type-insert (llrb-type table))
       (root table)
       key
       #f
       (if update
	   (lambda (n)
	     (let ((v (update (binding-node-value n))))
	       (set! result v)
	       (make-binding-node #f #f #f (binding-node-key n) v)))
	   (lambda (n) (set! result (binding-node-value n)) n))
       (lambda ()
	 (if (and (pair? default) (procedure? (car default)))
	     (let ((thunk (car default))) (%make-new-binding-node key (thunk)))
	     (error "generic-table-update! no default" default)))))
     result))

 (: generic-table-fold ((procedure (* * :table:) *) * :table: -> *))
 (define (generic-table-fold proc init table)
   (check-table table 'generic-table-fold)
   (ensure procedure? proc)
   ((llrb-tree-type-fold (llrb-type table))
    (lambda (node init) (proc (binding-node-key node) (binding-node-value node) init))
    init (root table)))

 (: generic-table-for-each ((procedure (* *) *) :table: -> *))
 (define (generic-table-for-each proc table)
   (check-table table 'generic-table-for-each)
   (ensure procedure? proc)
   ((llrb-tree-type-for-each (llrb-type table))
    (lambda (node) (proc (binding-node-key node) (binding-node-value node)))
    (root table))
   #f)

 (: generic-table/min (:table: (procedure () * *) --> * *))
 (define (generic-table/min table default)
   (check-table table 'generic-table/min)
   (let ((node ((llrb-tree-type-min (llrb-type table)) (root table))))
     (if node (values (binding-node-key node) (binding-node-value node))
	 (begin
	   (ensure procedure? default)
	   (default)))))

 (: generic-table/delete-min! (:table: (procedure () * *) -> * *))
 (define (generic-table/delete-min! table default)
   (if (generic-table-empty? table)
       (begin
	 (ensure procedure? default)
	 (default))
       ((llrb-tree-type-delete-min (llrb-type table))
	(root table)
	(lambda (root node)
	  (root-set! table root)
	  (values (binding-node-key node) (binding-node-value node))))))

 )

;; This mutating version will go away.  It is left here for now to
;; show how bad a difference the mutations are.
(include "fixnum-table-mutating.scm")

(module
 llrb-tree
 *
 (import scheme chicken)
 (import (prefix llrb-m-fixnum-table mu:))
 ;; This is a bit unfortune as it doubles several prefixes.
 ;;
 ;; At the other hand removing those in the source makes it harder to
 ;; maintain.  Therfore one should better import those specialized
 ;; versions directly if used. 
 (reexport (prefix llrb-string-table string-)
	   (prefix llrb-fixnum-table fixnum-)
	   (prefix llrb-symbol-tree symbol-)
	   llrb-generic-tree)
 (define str2sym string-str2sym)
 )
