(module
 llrb-m-string-table
 (
  make-string-table
  string-table-empty?
  string-table-ref
  string-table-ref/default
  string-table-set!
  string-table-delete!
  string-table-for-each
  string-table-fold
  string-table/min string-table/delete-min!
  )
 (import scheme chicken)
 (import llrb-syntax)
 ;; (include "llrbsyn.scm")

 ;; ordered allocation free, impure string llrb tree

 (cond-expand
  (never
   (define-type :llrb-node: (struct <llrb-node>))
   (define-record-type <llrb-node>
     (make-llrb-node color left right key value)
     llrb-node?
     (color llrb-node-color llrb-node-color-set!)
     (left llrb-node-left llrb-node-left-set!)
     (right llrb-node-right llrb-node-right-set!)
     (key llrb-node-key llrb-node-key-set!)
     (value llrb-node-value llrb-node-value-set!)))
  (else
   (define-type :llrb-node: (vector * * * * *))
   (define-inline (make-llrb-node color left right key value)
     (##sys#vector color left right key value))
   (define-inline (llrb-node-color n) (##sys#slot n 0))
   (define-inline (llrb-node-color-set! n v) (##sys#setslot n 0 v))
   (define-inline (llrb-node-left n) (##sys#slot n 1))
   (define-inline (llrb-node-left-set! n v) (##sys#setslot n 1 v))
   (define-inline (llrb-node-right n) (##sys#slot n 2))
   (define-inline (llrb-node-right-set! n v) (##sys#setslot n 2 v))
   (define-inline (llrb-node-key n) (##sys#slot n 3))
   (define-inline (llrb-node-key-set! n v) (##sys#setslot n 3 v))
   (define-inline (llrb-node-value n) (##sys#slot n 4))
   (define-inline (llrb-node-value-set! n v) (##sys#setslot n 4 v))))


 (define-type :table: (struct <llrb-string-table>))
 (define-record-type <llrb-string-table>
   (%make-string-table root)
   string-table?
   (root root))
 
 ;; (: string-ordered-llrbtree-node-insert! ((struct <llrb-node>) string * -> (struct <llrb-node>)))

 (define-syntax llrb-node-update
   (syntax-rules (left: right: color:)
     ((_ 1 n l r c ())
      (make-llrb-node c l r (llrb-node-key n) (llrb-node-value n)))
     ((_ 1 n l r c (left: v . more))
      (llrb-node-update 1 n v r c more))
     ((_ 1 n l r c (right: v . more))
      (llrb-node-update 1 n l v c more))
     ((_ 1 n l r c (color: v . more))
      (llrb-node-update 1 n l r v more))
     ((_ n . more)
      (llrb-node-update
       1 n (llrb-node-left n) (llrb-node-right n) (llrb-node-color n) more))
     ))

 (define-syntax llrb-node-update!
   (syntax-rules (left: right: color:)
     ((_ n) n)
     ((_ n left: v . more)
      (begin
	(llrb-node-left-set! n v)
	(llrb-node-update n . more)))
     ((_ n right: v . more)
      (begin
	(llrb-node-right-set! n v)
	(llrb-node-update n . more)))
     ((_ n color: v . more)
      (begin
	(llrb-node-color-set! n v)
	(llrb-node-update n . more)))
     ))

 (define-syntax %string=?
   (syntax-rules ()
     ((_ s1 si2)
      (let ((s2 si2))
	;; (##sys#check-string s1 'string-aggregate)
	;; (##sys#check-string s2 'string-aggregate)
	(string=? s1 s2)))))

 (define-syntax %string<?
   (syntax-rules ()
     ((_ si1 si2)
      #;(let ((s1 si1) (s2 si2))
	;;       (##sys#check-string s1 'string-aggregate)
	;;       (##sys#check-string s2 'string-aggregate)
	(let ((len1 (##core#inline "C_block_size" s1))
	      (len2 (##core#inline "C_block_size" s2)) )
	  (cond
	   ((fx= len1 len2)
	    (##core#inline "AC_string_compare" s1 s2 len1 0))
	   ((##core#inline "AC_lessp" len1 len2)
	    (##core#inline "AC_string_compare" s1 s2 len1 1))
	   (else (##core#inline "AC_string_compare" s1 s2 len2 0))) ))
      (string<? si1 si2))))

 (define-syntax llrb-node-key-node-eq?
   (syntax-rules ()
     ((_ key node)
      ;; don't use just "string=?" - in chicken it will miss the (most common) a=b case
      ;; (Actually the above comment is ages old.  Is it still valid?)
      (or (eq? key (llrb-node-key node)) (%string=? key (llrb-node-key node))))))

 (define-syntax llrb-node-key-node-lt
   (syntax-rules () ((_ key node) (%string<? key (llrb-node-key node)))))

 (define-syntax llrb-node-node-node-=?
   (syntax-rules ()
     ((_ node1 node2) (%string=? (llrb-node-key node1) (llrb-node-key node2)))))

 (define-syntax llrb-node-node-node-lt
   (syntax-rules ()
     ((_ node1 node2) (%string<? (llrb-node-key node1) (llrb-node-key node2)))))

 (define-llrbtree/positional
   (ordered)
   llrb-node-update!
   llrb-node-init!	      ;; defined
   llrb-node-node-lookup      ;; defined
   llrb-node-min	      ;; defined
   llrb-node-node-fold	      ;; defined
   llrb-node-node-for-each    ;; defined
   llrb-node-node-insert!     ;; defined
   llrb-node-delete!	      ;; defined
   llrb-node-node-delete-min! ;; defined
   llrb-node-empty?	      ;; defined
   llrb-node-key-node-eq?
   llrb-node-node-node-=?
   llrb-node-key-node-lt
   llrb-node-node-node-lt     ;; before? node order
   llrb-node-left
   llrb-node-right
   llrb-node-color
   )

 (: make-string-table (--> :table:))

 (define (make-string-table)
   (%make-string-table (llrb-node-init! (make-llrb-node #f #f #f  #f #f))))

 (: string-table-empty? (:table: --> boolean))
 (define (string-table-empty? table)
   (ensure string-table? table)
   (llrb-node-empty? (root table) (root table)))

 (: string-table-delete! (:table: string -> *))
 (define (string-table-delete! table key)
   (ensure string? key)
   (ensure string-table? table)
   (llrb-node-delete! (root table) key))

 (: string-table-set! (:table: string * -> boolean))
 (define (string-table-set! table key value)
   (ensure string? key)
   (ensure string-table? table)
   (llrb-node-node-insert! (root table) key #f (make-llrb-node #f #f #f key value) #f)
   #t)

 (: string-table-ref/default (:table: string * --> *))
 (define (string-table-ref/default table key default)
   (ensure string? key)
   (ensure string-table? table)
   (let ((node (llrb-node-node-lookup (root table) key)))
     (if node (llrb-node-value node) default)))

 (: string-table-ref (:table: string (procedure () *) -> *))
 (define (string-table-ref table key default)
   (ensure string? key)
   (ensure string-table? table)
   (let ((node (llrb-node-node-lookup (root table) key)))
     (if node (llrb-node-value node) (default))))

 (: string-table-fold ((procedure (string * *) *) * :table: -> *))
 (define (string-table-fold proc init table)
   (ensure string-table? table)
   (ensure procedure? proc)
   (llrb-node-node-fold
    (lambda (node init) (proc (llrb-node-key node) (llrb-node-value node) init))
    init (root table)))

 (: string-table-for-each ((procedure (string *) *) :table: -> *))
 (define (string-table-for-each proc table)
   (ensure string-table? table)
   (ensure procedure? proc)
   (llrb-node-node-for-each
    (lambda (node) (proc (llrb-node-key node) (llrb-node-value node)))
    (root table))
   #f)

 (: string-table/min (:table: (procedure () string *) --> string *))
 (define (string-table/min table default)
   (ensure string-table? table)
   (let ((node (llrb-node-min (root table))))
     (if node (values (llrb-node-key node) (llrb-node-value node))
	 (begin
	   (ensure procedure? default)
	   (default)))))

 (: string-table/delete-min! (:table: (procedure () string *) -> string *))
 (define (string-table/delete-min! table default)
   (ensure string-table? table)
   (let ((node (llrb-node-node-delete-min! (root table))))
     (if node (values (llrb-node-key node) (llrb-node-value node))
	 (begin
	   (ensure procedure? default)
	   (default)))))

 
 )
